// code by jph

#ifndef HASH_INCLUDED
#define HASH_INCLUDED

static char HASH_BUFFER[1024];

template<class comp,class type>
class bucket {
  public:
    comp key;  // =, ==
    type data; // =
};

template<class comp,class type>
class hash {
  public:
    hash<comp,type>(void) {};
    hash<comp,type>(int size,int (*func)(comp));    //init with # of buckets and hash function (ptr)
    ~hash<comp,type>(void);
    int add(const comp& key,const type& data);
    int get(const comp& key,type &data);
    int has(const comp& key);
    double variance(void);
    char* toInfo(void);
    char* toString(void); 
//----------------------------------------------------------------------
    tensor<linked<bucket<comp,type>>*> table;
    int numel;
  private:
    int (*func)(comp);
};

template<class comp,class type>
hash<comp,type>::hash(int size,int (*func)(comp)) {
  table=tensor<linked<bucket<comp,type>>*>(1,&size);
  for (int c1=0;c1<table.numel;c1++)
    table(c1)=new linked<bucket<comp,type>>();
  this->func=func;
  numel=0;
}

template<class comp,class type>
hash<comp,type>::~hash(void) {
  for (int c1=0;c1<table.numel;c1++)
    delete table(c1);
}

template<class comp,class type>
int hash<comp,type>::add(const comp& key,const type& data) {
  int adr=func(key);
  bucket<comp,type> buck;
  while (table(adr)->token(buck))
    if (buck.key==key) {
      table(adr)->reset(); // reset tokenizer
      return 1;
    }
  buck.key=key;
  buck.data=data;
  table(adr)->add(buck); // TODO maybe 'pre' is generally better
  ++numel;
  return 0;
}

template<class comp,class type>
int hash<comp,type>::get(const comp& key,type &data) {
  int adr=func(key);
  bucket<comp,type> buck;
  while (table(adr)->token(buck))
    if (buck.key==key) {
      data=buck.data; // return existing data
      table(adr)->reset(); // reset tokenizer
      return 1;
    }
  return 0;
}

template<class comp,class type>
int hash<comp,type>::has(const comp& key) {
  int adr=func(key);
  bucket<comp,type> buck;
  while (table(adr)->token(buck))
    if (buck.key==key) {
      table(adr)->reset(); // reset tokenizer
      return 1;
    }
  return 0;
}

template<class comp,class type> 
double hash<comp,type>::variance(void) {
  double mean=numel/(double)table.numel;
  double sum=0;
  double tmp;
  for (int c1=0;c1<table.numel;c1++) {
    tmp=table(c1)->numel-mean;
    sum+=tmp*tmp;
  }
  return sum/(table.numel-1.0);
}

template<class comp,class type>
char* hash<comp,type>::toInfo(void) {
  sprintf(HASH_BUFFER,"%p hash<%i,%i>(%i/%i occ=%5.1f%% var=%5.1f)",this,sizeof(comp),sizeof(type),numel,table.numel,numel*100/(double)table.numel,variance());
  return HASH_BUFFER;
}

template<class comp,class type>
char* hash<comp,type>::toString(void) {
  sprintf(HASH_BUFFER,"%p hash<%i,%i>(%i/%i occ=%5.1f%% var=%5.1f:",this,sizeof(comp),sizeof(type),numel,table.numel,numel*100/(double)table.numel,variance());
  if (table.numel<20) {
    for (int c0=0;c0<table.numel;++c0)
      sprintf(&HASH_BUFFER[strlen(HASH_BUFFER)],"%i ",table(c0)->numel);
  }
  sprintf(&HASH_BUFFER[strlen(HASH_BUFFER)-1],")");
  return HASH_BUFFER;
}

//----------------------------------------------------------------------

#define  HASH_TOTEST 5

void hash_flush(const hash<int,float>& myHash,char *name) {
  FILE *fid=fopen(name,"w");
  if (fid) {
    bucket<int,float> buck;
    for (int c0=0;c0<myHash.table.numel;++c0)
      while (myHash.table.data[c0]->token(buck))
        fprintf(fid,"%i\t%i\t%f\n",c0,buck.key,buck.data);
    fclose(fid);
  }
}

void hash_flush(const hash<tensor<int>,tensor<float>>& myHash,char *name) {
  FILE *fid=fopen(name,"w");
  if (fid) {
    bucket<tensor<int>,tensor<float>> buck;
    for (int c0=0;c0<myHash.table.numel;++c0)
      while (myHash.table.data[c0]->token(buck)) {
        printf("%s\n",buck.data.toString());
        fprintf(fid,"[%i]\t\t",c0);
        fprintf(fid,"%s\t\t",buck.key.toString());
        fprintf(fid,"%s\n",buck.data.toString());
      }
    fclose(fid);
  }
}

#endif
