// code by jph adapted from Paul Kocher
// sha.hpp does not implement official sha(X) hash function

#ifndef SHA_INCLUDED
#define SHA_INCLUDED

#define SHA_RESET           word=0; \
                            code.as(0x67452301,0xefcdab89,0x98badcfe,0x10325476,0xc3d2e1f0); \
                            size=0; \
                            numel=0;

#define SHA_ROTATE(X,n)     ((X)<<n|(X)>>32-n)

#define SHA_SHIFT           hold.data[4]=hold.data[3]; \
                            hold.data[3]=hold.data[2]; \
                            hold.data[2]=SHA_ROTATE(hold.data[1],30); \
                            hold.data[1]=hold.data[0]; \
                            hold.data[0]=TEMP;

static char SHA_BUFFER[1024];

class sha {
  public:
    sha(void);
    void add(const void *data,int bytes); // len bytes of data
    tensor<int> flush(void);
    char* toInfo(void);
    char* toString(void);
    template<class type> static int hash(tensor<type> myTensor); // ok
//----------------------------------------------------------------------
    tensor<unsigned int> word; // 80 integer buffer to accumulate bytes that are flushed to code
    tensor<unsigned int> code; //  5 integer buffer to hold hash values
    int size;  // current position in 0...63, corresponding to the first 16 entries of word
    int numel; // counts total bytes added
};

sha::sha(void) {
  word=tensor<unsigned int>::vector(80); // 80*4=320 bytes
  code=tensor<unsigned int>::vector(5);  //  5*4= 20 bytes
  SHA_RESET
}

void sha::add(const void *data,int bytes) {
  const unsigned char* cast=(unsigned char*)data; // (!unsigned is important here!)
  int addr;
  for (int c1=0;c1<bytes;++c1) {
    addr=size>>2;
    word.data[addr]<<=8;
    word.data[addr]|=cast[c1];
    ++size%=64;   // wtf
    if (!size) {  // data only enters first 16 ints of word, 16*4=64 bytes written
      int c0;
      unsigned int TEMP;
      tensor<unsigned int> hold(code);
      for (c0=16;c0<80;++c0)
        word.data[c0]=SHA_ROTATE(word.data[c0-3]^word.data[c0-8]^word.data[c0-14]^word.data[c0-16],1);
      for (c0= 0;c0<20;++c0) { TEMP=SHA_ROTATE(hold.data[0],5)+(((hold.data[2]^hold.data[3])&hold.data[1])^hold.data[3])                +hold.data[4]+word.data[c0]+0x5a827999; SHA_SHIFT }
      for (c0=20;c0<40;++c0) { TEMP=SHA_ROTATE(hold.data[0],5)+(hold.data[1]^hold.data[2]^hold.data[3])                                 +hold.data[4]+word.data[c0]+0x6ed9eba1; SHA_SHIFT }
      for (c0=40;c0<60;++c0) { TEMP=SHA_ROTATE(hold.data[0],5)+((hold.data[1]&hold.data[2])|(hold.data[3]&(hold.data[1]|hold.data[2]))) +hold.data[4]+word.data[c0]+0x8f1bbcdc; SHA_SHIFT }
      for (c0=60;c0<80;++c0) { TEMP=SHA_ROTATE(hold.data[0],5)+(hold.data[1]^hold.data[2]^hold.data[3])                                 +hold.data[4]+word.data[c0]+0xca62c1d6; SHA_SHIFT }
      code+=hold;
    }
  }
  numel+=bytes;
}

tensor<int> sha::flush(void) {
  unsigned char pad_0x00=0x00;
  while (size!=60)
    add(&pad_0x00,1);  
  add(&numel,4);
  tensor<int> ret=code.to<int>();
  SHA_RESET
  return ret;
}

template<class type>
static int sha::hash(tensor<type> myTensor) { // ok
  sha mySha;
  mySha.add(myTensor.data,myTensor.numel*sizeof(type));
  return mySha.flush().sum()&0x7fffffff;
}

char* sha::toInfo(void) {
  sprintf(SHA_BUFFER,"sha(%d,%d)%s",numel,size,code.toString(" %08x"));
  return SHA_BUFFER;
}

char* sha::toString(void) {
  sprintf(SHA_BUFFER,"%s",word.toString(" %08x "));
  return SHA_BUFFER;
}

#endif
