// code by jph

class aniLink : public animated {
public:
  animated*       ani[2];   // animated objects
  tensor<double>  lev[2];   // vector 4   : point  (=> lev[i](3)=1) of joint in object coordinates (lev is constant over the course of the simulation)
  tensor<double>  axs[2];   // vector 4   : vector (=> axs[i](3)=0) of joint in object coordinates (axs is constant over the course of the simulation)
  tensor<double>  trf[2];   // matrix 4x4 : transformation in object coordinates to point of joint (to draw a cone to the joint)
  tensor<double>  shf[2];   // matrix 4x4 : transformation in object coordinates to point of joint (to draw a cylinder as the hinge)
  tensor<double>  cnL[2];   // matrix 4x4 : transformation in object coordinates to point of joint (to draw cones to the hinge)
  tensor<double>  cnR[2];   // matrix 4x4 : transformation in object coordinates to point of joint (to draw cones to the hinge)
  tensor<double>  ect;      // external control
  int             rank;     // rank=3 spherical, rank=5 hinge
  double          fmu;      // friction fmu>=0
  double          mot;      // motor
  
  aniLink(animated* ani0,tensor<double> lev0,
          animated* ani1,tensor<double> lev1,
          tensor<double> dir,double rev,
          double fmu) {
    setLink(0,ani0,lev0);
    setLink(1,ani1,lev1);      
    dir(3)=0;
    dir=dir.normal();
    ani[1]->pos.in(12,( ani[1]->pos[3] + ani[0]->pos*lev[0] - ani[1]->pos*lev[1] ).data,3);           // p1 == p1 + p0 + L0 - p1 - L1 // ok
    ani[1]->ome=ani[1]->pos.solve( ani[0]->pos*ani[0]->ome - dir*rev );                               // w1 == w0 - dir*rev   =>  w0 - w1 == dir*rev // ok
    ani[1]->vel=ani[0]->vel + ani[0]->pos*ani[0]->ome.x(lev[0]) - ani[1]->pos*ani[1]->ome.x(lev[1]);  // v1 == v0 + W0 x l0 - W1 x l1 // ok      
    axs[0]=tensor<double>(1,4);
    axs[1]=tensor<double>(1,4);
    ect=tensor<double>(1,4);      // this is for control
    this->fmu=fmu;
  }
  
  tensor<double> getRotation(void) {
    return ani[0]->pos*ani[0]->ome-ani[1]->pos*ani[1]->ome; // R0.w0 - R1.w1
  }
  
  void control(void) {
    for (int c0=0;c0<2;++c0) {
      ani[c0]->acc=0;
      ani[c0]->tor=0;
    }
    acc=0;
    tor=0;
    ect=getRotation()*(fmu*-1.0)+ani[0]->pos*axs[0]*mot; // world coords
  }
  
  tensor<double>* control_lhs(aniLink* lnk1) {
    tensor<double> ret(2,6,6);
    if (ani[0]==lnk1->ani[0]) ret+=control_lhs(0,lnk1,0);
    if (ani[1]==lnk1->ani[1]) ret+=control_lhs(1,lnk1,1);
    if (ani[0]==lnk1->ani[1]) ret-=control_lhs(0,lnk1,1);
    if (ani[1]==lnk1->ani[0]) ret-=control_lhs(1,lnk1,0);
    return new tensor<double>(ret);
  }
  
  tensor<double> control_lhs(int ind0,aniLink* lnk1,int ind1) {
    double fac=1/ani[ind0]->mass;
    return tensor<double>(tensor<tensor<double>*>(2,2,2).as(
      new tensor<double>(ani[ind0]->link_lhs(lev[ind0],lnk1->lev[ind1])-tensor<double>(2,3,3).diag(fac,fac,fac)),
      new tensor<double>(ani[ind0]->link_lhs(axs[ind0],lnk1->lev[ind1])),
      new tensor<double>(ani[ind0]->link_lhs(lev[ind0],lnk1->axs[ind1])),
      new tensor<double>(ani[ind0]->link_lhs(axs[ind0],lnk1->axs[ind1]))));
  }

  tensor<double>* control_rhs(void) {
    return new tensor<double>(tensor<tensor<double>*>(1,2).as(
      new tensor<double>( ani[0]->link_rhs(lev[0])-ani[1]->link_rhs(lev[1]) ),
      new tensor<double>( ani[0]->link_rhs(axs[0])-ani[1]->link_rhs(axs[1]) )));    
  }
  
  tensor<double> control_ect(int ind0,aniLink* lnk1,int ind1) {
    return tensor<double>(tensor<tensor<double>*>(2,2,1).as(
      new tensor<double>( ani[ind0]->link_ect(lev[ind0]) ),
      new tensor<double>( ani[ind0]->link_ect(axs[ind0]) )));
  }
  
  tensor<double> control_ect(aniLink* lnk1) {
    tensor<double> ret(2,6,3);
    if (ani[0]==lnk1->ani[0]) ret+=control_ect(0,lnk1,0);
    if (ani[1]==lnk1->ani[1]) ret+=control_ect(1,lnk1,1);
    if (ani[0]==lnk1->ani[1]) ret-=control_ect(0,lnk1,1);
    if (ani[1]==lnk1->ani[0]) ret-=control_ect(1,lnk1,0);
    return ret;
  }

  void intake(double* sol) {
    tensor<double> sgn=tensor<double>(1,2).as( 1.0,-1.0);
    acc.in(0, sol   ,3);
    tor.in(0,&sol[3],3);
    for (int c0=0;c0<2;++c0) {
      ani[c0]->acc+=ani[c0]->pos.solve(acc*(sgn(c0)/ani[c0]->mass));
      ani[c0]->tor+=lev[c0].x(ani[c0]->pos.solve(acc*sgn(c0)))+
                    axs[c0].x(ani[c0]->pos.solve(tor*sgn(c0)))+
                              ani[c0]->pos.solve(ect*sgn(c0));
    }
  }
  
  // constructor taking in single world coordinate
  int isConsistent(void) {
    double prec=tensor<double>::precision=1e-4;
    assert( ani[0]!=ani[1] );
    assert( ani[0]->pos*lev[0] ==
            ani[1]->pos*lev[1] );                             // position of hinge    
    assert( ani[0]->vel+ani[0]->pos*ani[0]->ome.x(lev[0]) ==
            ani[1]->vel+ani[1]->pos*ani[1]->ome.x(lev[1]) );  // linear velocity of hinge
    assert( ani[0]->pos*axs[0] ==
            ani[1]->pos*axs[1] );                             // direction of hinge ok
    assert( ani[0]->pos*ani[0]->ome.x(axs[0]) ==
            ani[1]->pos*ani[1]->ome.x(axs[1]) );              // dt of direction of hinge ok
    assert( abs(lev[0](3)-1)<prec &&
            abs(lev[1](3)-1)<prec );
    assert( abs(axs[0](3))<prec &&
            abs(axs[1](3))<prec );
    return 1;
  }

  void draw_dyn(void) {
    double alp=0.8;
    double wdt=0.05;
    glPushAttrib(GL_DEPTH_TEST);
      glDisable(GL_DEPTH_TEST);
      glPushMatrix();
        glColor4d( 1.0, 1.0, 0.0, alp);
        glColor4d( 1.0, 0.8, 0.3, alp);
        glMultMatrixd(glMatrixFromVector(acc,ani[0]->pos*lev[0]).data);
        glutSolidCylinder(0,acc.norm(2),wdt,10);
      glPopMatrix();
      glPushMatrix();
        glColor4d( 0.8, 0.3, 1.0, alp);
        glMultMatrixd(glMatrixFromVector(tor,ani[0]->pos*lev[0]).data);
        glutSolidCylinder(0,tor.norm(2),wdt,10);
      glPopMatrix();
      glPushMatrix();
        glColor4d( 0.3, 1.0, 0.8, alp);
        glMultMatrixd(glMatrixFromVector(ect,ani[0]->pos*lev[0]).data);
        glutSolidCylinder(0,ect.norm(2),wdt,10);
      glPopMatrix();
    glPopAttrib();
    if (keyboard('e')) 
      defects(); // adds errors to bulletin    
  }

  void defects(void) {
    tensor<double> dp=ani[0]->pos*lev[0] - ani[1]->pos*lev[1];
    tensor<double> dv=ani[0]->vel+ani[0]->pos*ani[0]->ome.x(lev[0]) - (ani[1]->vel+ani[1]->pos*ani[1]->ome.x(lev[1]));
    tensor<double> da=ani[0]->pos*axs[0] - ani[1]->pos*axs[1];
    tensor<double> dw=ani[0]->pos*ani[0]->ome.x(axs[0]) - ani[1]->pos*ani[1]->ome.x(axs[1]);
    opengl_bulletin.add(tensor<char>::string("dp ")&tensor<char>::string( dp.toString() ));
    opengl_bulletin.add(tensor<char>::string("dv ")&tensor<char>::string( dv.toString() ));
    opengl_bulletin.add(tensor<char>::string("da ")&tensor<char>::string( da.toString() ));
    opengl_bulletin.add(tensor<char>::string("dw ")&tensor<char>::string( dw.toString() ));
  }

  void correct(void) {
    // TODO find rotation to map   z1 to z0  => apply it to obtain   R0.z0 == R1.z1
    ani[1]->pos.in(12,( ani[1]->pos[3] + ani[0]->pos*lev[0] - ani[1]->pos*lev[1] ).data,3);           // p1 == p1 + p0 + L0 - p1 - L1 // ok
    // TODO update ome
    ani[1]->vel=ani[0]->vel + ani[0]->pos*ani[0]->ome.x(lev[0]) - ani[1]->pos*ani[1]->ome.x(lev[1]);  // v1 == v0 + W0 x l0 - W1 x l1 // ok      
  }

protected:  
  void setLink(int ind,animated* ani,tensor<double> lev) { // called by constructors
    lev(3)=1;
    this->ani[ind]=ani;
    this->lev[ind]=lev;
    trf[ind]=glMatrixFromVector(lev,tensor<double>(1,3));
  }

};

#include "aniLink1.cpp" // hinges
#include "aniLink3.cpp" // spherical joints
