// code by jph

#define _CRT_SECURE_NO_DEPRECATE  // no warnings when using std::c in vcpp compiler 14

// #define NDEBUG
// #define OPENGL_FULLSCREEN "1680x1050:32"
// #define OPENGL_FULLSCREEN "800x600:32"

#include "../_util/_opengl.hpp"
#include "animated.hpp"
#include "aniLink.cpp"
#include "aniSkeleton.cpp"
#include "aniAvatar.cpp"
#include "aniLight.cpp"
#include "aniCamera.cpp"

aniAvatar* ava;
aniCamera cam;
aniSkeleton* skl=0;
linked<animated*> linked_ani;

int recindicate=0;

int FACTOR=10;

#define EQM    0.288675*3
#define EQH    0.577350*3

double msec=0;
bool displayskeleton=true;
aniLink1* aniMotor;
aniLink1* aniEngin;
aniLink3* aniWobbl;
double clkMotor;
tensor<double> wob;
tensor<double> xrt;

void configuration(int type) { // defines body and collision characteristics
  int c0;
  aniMotor=aniEngin=0;
  aniWobbl=0;
  clkMotor=0;
  wob=tensor<double>(1,4).as( 1.0, 0.0, 0.0, 0.0);
  xrt=tensor<double>::randn(1,3).normal();
  if (skl) {
    skl->rmv=opengl_time;
    tensor<animated*> bdy=skl->whos();
    for (c0=0;c0<bdy.numel;c0++)
      bdy(c0)->rmv=opengl_time;    
  }
  skl=new aniSkeleton();
  tensor<animated*> tensor_ani;
  switch (type) {
    case 1:
      tensor_ani=tensor<animated*>::vector(2).as(new aniAvatar(),new aniAvatar());
    	tensor_ani(0)->col=opengl_color_red;
    	tensor_ani(1)->col=opengl_color_cyan;
    	tensor_ani(0)->ome=tensor<double>::vector(4).as( 1.0, 1.3, 0.4, 0.0);
    	tensor_ani(0)->setEllipsoid(tensor<double>::rand(1,4)*0.5+0.4);
    	tensor_ani(1)->setEllipsoid(tensor<double>::rand(1,4)*0.5+0.4);
    	skl->link.add(aniMotor=new aniLink1(tensor_ani(0),tensor<double>::vector(4).as( 1.0, 1.5, 0.2, 1.0),
                                 tensor_ani(1),tensor<double>::vector(4).as(-2.0, 0.3,-0.3, 1.0),
                                 tensor<double>::randn(1,4),
                                 2.0f,
                                 0.2f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
      break;
    case 2: {
      tensor_ani=tensor<animated*>::vector(2).as(new aniAvatar(),new aniAvatar());
    	tensor_ani(0)->col=opengl_color_red;
    	tensor_ani(1)->col=opengl_color_cyan;
      tensor_ani(0)->ome=tensor<double>::vector(4).as( 1.0, 1.3, 0.4, 0.0);
    	tensor_ani(0)->setEllipsoid(tensor<double>::rand(1,4)*0.5+0.4);
    	tensor_ani(1)->setEllipsoid(tensor<double>::rand(1,4)*0.5+0.4);
    	tensor<double> lv0=tensor<double>::rand(1,4)*2; lv0(3)=1;
    	skl->link.add(aniWobbl=new aniLink3(tensor_ani(0),lv0,
                                 tensor_ani(1),tensor<double>::vector(4).as(-2.0, 0.3,-1.0, 1.0),
                                 //tensor<double>::rand(1,4),
                                 lv0*(-1.0),
                                 3.0f,
                                 0.5f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
      break;
  	}
    case 3: {
      tensor_ani=tensor<animated*>::vector(2).as(new aniAvatar(),new aniAvatar());
    	tensor_ani(0)->col=opengl_color_red;
    	tensor_ani(1)->col=opengl_color_cyan;
      tensor_ani(0)->ome=tensor<double>::vector(4).as( 0.0, 0.0, 2.9, 0.0);
    	tensor_ani(0)->setEllipsoid(tensor<double>::rand(1,4)*0.5+0.4);
    	tensor_ani(1)->setEllipsoid(tensor<double>::rand(1,4)*0.5+0.4);
    	tensor<double> lv0=tensor<double>::vector(4).as( 1.1, 0.5, 0.0, 1.0);
    	tensor<double> lv1=lv0*(-1.0);
    	skl->link.add(aniMotor=new aniLink1(tensor_ani(0),lv0,
                                 tensor_ani(1),lv1,
                                 lv1,
                                 6.0f,
                                 0.2f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
      break;
  	}
    case 4: {
      tensor_ani=tensor<animated*>::vector(3).as(new aniAvatar(),new aniAvatar(),new aniAvatar());
    	tensor_ani(0)->col=opengl_color_red;
    	tensor_ani(1)->col=opengl_color_green;
    	tensor_ani(2)->col=opengl_color_blue;
      tensor_ani(0)->ome=tensor<double>::vector(4).as( 0.0, 0.0, 0.0, 0.0);
    	tensor_ani(0)->setEllipsoid(tensor<double>::vector(4).as( 0.3, 1.0, 1.0, 1.0));
    	tensor_ani(1)->setEllipsoid(tensor<double>::vector(4).as( 0.1, 1.0, 0.2, 1.0));
    	tensor_ani(2)->setEllipsoid(tensor<double>::vector(4).as( 0.1, 0.2, 1.0, 1.0));
    	tensor<double> acx=tensor<double>::vector(4).as( 0.6, 0.0, 0.0, 1.0);
    	tensor<double> lv0=tensor<double>::vector(4).as( 0.6, 0.8, 0.0, 1.0);
    	tensor<double> lv1=lv0*(-1.0);
    	tensor<double> lv2=tensor<double>::vector(4).as(-0.6,-0.7, 0.7, 1.0);
    	tensor<double> lv3=lv2*(-1.0);
    	skl->link.add(aniMotor=new aniLink1(tensor_ani(0),lv0,
                                 tensor_ani(1),lv1,
                                 acx,
                                 -2.0f,
                                 0.5f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
    	skl->link.add(new aniLink1(tensor_ani(0),lv2,
                                 tensor_ani(2),lv3,
                                 acx,
                                 3.0f,
                                 0.0f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
      break;
  	}
    case 5: {
      tensor_ani=tensor<animated*>::vector(3).as(new aniAvatar(),new aniAvatar(),new aniAvatar()); // blu
    	tensor_ani(0)->col=opengl_color_red;
    	tensor_ani(1)->col=opengl_color_green;
    	tensor_ani(2)->col=opengl_color_cyan;
    	tensor_ani(0)->pos=glD(tensor<double>::vector(3).as( 0.9, 0.4, 0.1)).exp();
    	tensor_ani(1)->pos=glD(tensor<double>::vector(3).as( 1.9,-0.4, 0.3)).exp();
    	tensor_ani(2)->pos=glD(tensor<double>::vector(3).as(-0.3, 0.2,-0.3)).exp();
      tensor_ani(0)->ome=tensor<double>::vector(4).as( 1.0, 1.3, 0.4, 0.0);
    	tensor_ani(0)->setEllipsoid(tensor<double>::vector(4).as( 0.7, 1.0, 0.3, 0.0));
    	tensor_ani(1)->setEllipsoid(tensor<double>::vector(4).as( 0.6, 0.5, 0.4, 0.0));
    	tensor_ani(2)->setEllipsoid(tensor<double>::vector(4).as( 0.4, 0.8, 0.8, 0.0));
    	skl->link.add(aniMotor=new aniLink1(tensor_ani(0),tensor<double>::vector(4).as( 1.0, 1.5, 0.7, 1.0),    	                                         
                                 tensor_ani(1),tensor<double>::vector(4).as(-2.0, 0.3,-1.0, 1.0),
                                 tensor<double>::randn(1,4),
                                 2.0f,
                                 0.5f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
    	skl->link.add(new aniLink3(tensor_ani(1),tensor<double>::vector(4).as( 1.0, 0.8,-0.7, 1.0),    	                                         
                                 tensor_ani(2),tensor<double>::vector(4).as(-1.0, 0.4, 0.3, 1.0),
                                 tensor<double>::randn(1,4),
                                 3.0f,
                                 0.5f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
      break;
  	}
    case 6: { // mercedes
      tensor_ani=tensor<animated*>::vector(4).as(new aniAvatar(),new aniAvatar(),new aniAvatar(),new aniAvatar()); // blu
    	tensor_ani(0)->col=opengl_color_red;
    	tensor_ani(1)->col=opengl_color_green;
    	tensor_ani(2)->col=opengl_color_blue;
    	tensor_ani(3)->col=opengl_color_cyan;
      tensor_ani(0)->ome=tensor<double>::vector(4).as( 0.0, 0.0, 1.0, 0.0);
    	tensor_ani(0)->setEllipsoid(tensor<double>::vector(4).as( 0.8, 0.8, 0.8, 0.0));
    	tensor_ani(1)->setEllipsoid(tensor<double>::vector(4).as( 0.2, 0.6, 0.6, 0.0));
    	tensor_ani(2)->setEllipsoid(tensor<double>::vector(4).as( 0.8, 0.2, 0.8, 0.0));
    	tensor_ani(3)->setEllipsoid(tensor<double>::vector(4).as( 0.7, 0.3, 0.3, 0.0));
    	skl->link.add(aniMotor=new aniLink1(tensor_ani(0),tensor<double>::vector(4).as( 1.2, 0.0, 0.0, 1.0),    	                                         
                                 tensor_ani(1),tensor<double>::vector(4).as(-0.6, 0.0, 0.0, 1.0),
                                 tensor<double>::vector(4).as(-1.0, 0.0, 0.0, 0.0),
                                 2.0f,
                                 0.05f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
    	skl->link.add(aniEngin=new aniLink1(tensor_ani(0),tensor<double>::vector(4).as( 0.0, 1.3, 0.0, 1.0),    	                                         
                                 tensor_ani(2),tensor<double>::vector(4).as( 0.0,-0.6, 0.0, 1.0),
                                 tensor<double>::vector(4).as( 0.0,-0.9, 0.0, 0.0),
                                 3.0f,
                                 0.0f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
    	skl->link.add(new aniLink3(tensor_ani(0),tensor<double>::vector(4).as(-0.7,-0.8, 0.5, 1.0),    	                                         
                                 tensor_ani(3),tensor<double>::vector(4).as( 0.7, 0.5,-0.2, 1.0),
                                 tensor<double>::randn(1,4),
                                 1.0f,
                                 0.1f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
      break;
  	}
    case 7: {
      tensor_ani=tensor<animated*>::vector(5).as(new aniAvatar(),new aniAvatar(),new aniAvatar(),new aniAvatar(),new aniAvatar()); // blu
    	tensor_ani(0)->col=opengl_color_red;  
    	tensor_ani(1)->col=opengl_color_green;
    	tensor_ani(2)->col=opengl_color_blue; 
    	tensor_ani(3)->col=opengl_color_cyan; 
    	tensor_ani(4)->col=opengl_color_yellow;
    	tensor_ani(0)->pos=glD(tensor<double>::vector(3).as( 0.9, 0.4, 0.1)).exp();
    	tensor_ani(1)->pos=glD(tensor<double>::vector(3).as( 1.9,-0.4, 0.3)).exp();
    	tensor_ani(2)->pos=glD(tensor<double>::vector(3).as(-0.3, 0.2,-0.3)).exp();
    	tensor_ani(3)->pos=glD(tensor<double>::vector(3).as( 0.1, 0.1, 0.1)).exp();
    	tensor_ani(4)->pos=glD(tensor<double>::vector(3).as( 0.1, 0.1, 0.1)).exp();
      tensor_ani(0)->setEllipsoid(tensor<double>::rand(1,4)*0.3+0.4);
    	tensor_ani(1)->setEllipsoid(tensor<double>::rand(1,4)*0.3+0.4);
    	tensor_ani(2)->setEllipsoid(tensor<double>::rand(1,4)*0.3+0.4);
    	tensor_ani(3)->setEllipsoid(tensor<double>::rand(1,4)*0.3+0.4);
    	tensor_ani(4)->setEllipsoid(tensor<double>::rand(1,4)*0.3+0.4);
    	skl->link.add(new aniLink3(tensor_ani(0),tensor<double>::randn(1,4),
                                 tensor_ani(1),tensor<double>::randn(1,4),
                                 tensor<double>::randn(1,4),
                                 0.0f,
                                 0.1f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
    	skl->link.add(aniWobbl=new aniLink3(tensor_ani(1),tensor<double>::randn(1,4),
                                 tensor_ani(2),tensor<double>::randn(1,4),
                                 tensor<double>::randn(1,4),
                                 2.0f,
                                 0.2f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
    	skl->link.add(aniMotor=new aniLink1(tensor_ani(2),tensor<double>::randn(1,4),
                                 tensor_ani(3),tensor<double>::randn(1,4),
                                 tensor<double>::randn(1,4),
                                 1.0f,
                                 0.5f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
    	skl->link.add(aniEngin=new aniLink1(tensor_ani(1),tensor<double>::randn(1,4),
                                 tensor_ani(4),tensor<double>::randn(1,4),
                                 tensor<double>::randn(1,4),
                                 1.0f,
                                 0.5f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
      break;
  	}
    case 8: {
      tensor_ani=tensor<animated*>::vector(3).as(new aniAvatar(),new aniAvatar(),new aniAvatar());
    	tensor_ani(0)->col=tensor<double>::vector(4).as( 0.9, 0.0, 0.0, 1.0);
    	tensor_ani(1)->col=tensor<double>::vector(4).as( 0.0, 0.9, 0.0, 1.0);
    	tensor_ani(2)->col=tensor<double>::vector(4).as( 0.0, 0.8, 0.8, 1.0);
    	double inr=0.6;
      tensor_ani(0)->setEllipsoid(tensor<double>::vector(4).as( 0.8, 0.8, 0.8, 1.0));
    	tensor_ani(1)->setEllipsoid(tensor<double>::vector(4).as( inr, inr, inr, 1.0));
    	tensor_ani(2)->setEllipsoid(tensor<double>::vector(4).as( inr, inr, inr, 1.0));
    	skl->link.add(aniMotor=new aniLink1(tensor_ani(0),tensor<double>::vector(4).as( 2.0, 0.0, 0.0, 1.0),
                                 tensor_ani(1),tensor<double>::vector(4).as(-0.5,-0.5, 0.0, 1.0),
                                 tensor<double>::vector(4).as( 0.0, 0.0, 1.0, 1.0),
                                 -2.0f,
                                 0.5f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
    	skl->link.add(new aniLink3(tensor_ani(0),tensor<double>::vector(4).as(-1.4, 0.0, 0.0, 1.0),
                                 tensor_ani(2),tensor<double>::vector(4).as( 0.0, 0.0, 1.1, 1.0),
                                 tensor<double>::vector(4).as( 0.0, 1.0, 0.0, 1.0),
                                 0.0f,
                                 0.2f)); // 1 -> very accurate, 10 -> not so accurate, but fun to watch
      break;
  	}
  }
  skl->normalize();
  skl->isConsistent();
  linked_ani.add(tensor_ani);
  linked_ani.add((animated*)skl); // update of "position" after objects
}

aniLight* li0;
aniLight* li1;

int main(int argc,char** argv) {
  opengl_recording_numel=70;
  opengl_recording_frame=10;   // take shot every x frames
#ifdef NDEBUG   // release mode
  opengl_size.as(230,153);    // video format for most of my animations (confirmed 20091010)
  opengl_position.as(100,100);
#else           // debug mode
  opengl_size.as(640,480);
  opengl_position.as(1000,30);
  opengl_position.as(100,100);
#endif
  opengl_title=tensor<char>::string("Animation of skeletons");
  opengl_main(argc,argv);

  ava=new aniAvatar();
  cam.carrier=ava;
  ava->pos(2,3)=5;
  ava->fva=tensor<double>(1,4).as(.3,.3,.3,0);
  ava->fot=tensor<double>(1,4).as(.3,.3,.3,0);
  ava->status=ANI_HUMANIZED;
  
  li0=new aniLight(0);
  li0->pos(2,3)=10;    
  linked_ani.add(li0);

	configuration(5);
	
  glClearColor( 0.2, 0.2, 0.2, 0.0);
  glutMainLoop();
  return 0;
}

void draw_all(void) {
  animated* ani;
  while (linked_ani.token(ani))
    ani->draw_std();
}

void display(void) {
	int c0,c1,c2;
	char glt[80];
  animated::dt=opengl_rate.s();
  ava->control();
  ava->motion();  
  li0->pos=ava->pos; // attach light0 to viewer
  opengl_bulletin.add(tensor<char>::string( "press keys 1,2,3,... to load skeleton" ));
  opengl_bulletin.add(tensor<char>::string( "hold SPACE to animate, hold e for errors" ));
  if (keyboard('y')) {
    glClearColor( 1, 1, 1, 0.0);
  } else {
    glClearColor( 0.2, 0.2, 0.2, 0.0);
    opengl_grid_xy();
  }

  glMultMatrixd(glI(cam.carrier->pos).data);
  
  aniLight::ambience();
  
  for (c0=1;c0<9;++c0)
    if (keyboard_down(48+c0))
      configuration(c0);
  
  opengl_bulletin.add(tensor<char>::string("M_lin ")&tensor<char>::string( skl->momentum_lin().toString() ));
  opengl_bulletin.add(tensor<char>::string("M_ang ")&tensor<char>::string( skl->momentum_ang().toString() ));
  opengl_bulletin.add(tensor<char>::string("E_kin ")&tensor<double>::string( skl->energy_kin() ));

  animated::dt=0.0005;
  if ((opengl_recording||keyboard(' '))&&skl) {
    for (c1=0;c1<FACTOR;c1++) {
      clkMotor+=animated::dt;
      if (aniMotor)
        aniMotor->mot=sin(clkMotor*3)*2;
      if (aniEngin) 
        aniEngin->mot=cos(clkMotor*2.2)*3;      
        
      skl->control();
      
      if (aniWobbl) {
        wob=glD(xrt*0.001).exp()*wob;
        aniWobbl->ect+=wob;
      }
      
      skl->motion();
      skl->correct();
    }
  }
  draw_all();
  
	animated* animated_;
  while (linked_ani.token(animated_)) {
    if (animated_->rmv.raw&&animated_->rmv<=opengl_time) {
      linked_ani.remove();
      delete animated_;
    }
  }
}
